import pytest
from typing import Union, List, Tuple
try:
    from dataclasses import dataclass, field, KW_ONLY
except ImportError:
    from dataclasses import dataclass, field
    KW_ONLY = None
from pydantic import ValidationError
from scityping import Serializable, Slice, Dataclass, config
from scityping.base import deep_reduce
from scityping.pydantic import BaseModel, dataclass as pydataclass

config.safe_packages.add(__name__)

# Dataclass definitions need to be in global scope in order to be deserializable
@dataclass
class DcStdTypes:
    i: int
    s: Tuple[str,str]
    _: KW_ONLY = None  # None allows to work with < 3.10
    f: Union[float,List[float]]=None
    slc: Slice=None
    _private: int = field(init=False)  # Test that fields excluded from init are also excluded from serialization

@dataclass
class Obj1:
    slc: Slice

@dataclass
class Obj1b:
    slc: Slice

def test_dataclass_and_pydantic():
    """
    Tests verifying the interoperability of dataclasses, Pydantic BaseModel and Serializable.
    """
    @dataclass
    class Obj2(Serializable):
        slc: Slice
        @dataclass
        class Data:
            slc: Slice
            def encode(obj):
                return (obj.slc,)

    @pydataclass
    class Obj2b(Serializable):
        slc: Slice
        @dataclass
        class Data:
            slc: Slice
            def encode(obj):
                return (obj.slc,)

    @pydataclass
    class Obj2c(Serializable):
        slc: Slice
        @pydataclass
        class Data:
            slc: Slice
            def encode(obj):
                return (obj.slc,)

    class Obj3(BaseModel):
        slc: Slice

    class Obj4(Serializable):
        slc: Slice
        def __init__(self, slc: slice):
            self.slc = slc
        @dataclass
        class Data:
            slc: Slice
            def encode(obj): return (obj.slc,)

    class Obj4b(Serializable):
        slc: Slice
        def __init__(self, slc: slice):
            self.slc = slc
        @pydataclass
        class Data:
            slc: Slice
            def encode(obj): return (obj.slc,)

    class Foo1(BaseModel):
        obj: Obj1
    class Foo2(BaseModel):
        obj: Obj2
    class Foo2b(BaseModel):
        obj: Obj2b
    class Foo2c(BaseModel):
        obj: Obj2c
    class Foo3(BaseModel):
        obj: Obj3
    class Foo4(BaseModel):
        obj: Obj4
    class Foo4b(BaseModel):
        obj: Obj4b
    class Foo5(BaseModel):   # Use `Dataclass` as an ABC:
        obj: Dataclass       # accepts any dataclass
    class Foo5b(BaseModel):
        obj: Obj1b          # Accepts only dataclasses which are subclasses of Obj1b

    foo1 = Foo1(obj=Obj1(slc=slice(3, 8)))
    foo2 = Foo2(obj=Obj2(slc=slice(3, 8)))
    foo2b = Foo2b(obj=Obj2b(slc=slice(3, 8)))
    foo2c = Foo2c(obj=Obj2c(slc=slice(3, 8)))
    foo3 = Foo3(obj=Obj3(slc=slice(3, 8)))
    foo4 = Foo4(obj=Obj4(slc=slice(3, 8)))
    foo4b = Foo4b(obj=Obj4b(slc=slice(3, 8)))
    foo5 = Foo5(obj=Obj1(slc=slice(3, 8)))
    foo5b = Foo5b(obj=Obj1b(slc=slice(3, 8)))
    with pytest.raises(ValidationError):  # Foo5b only accepts dataclasses of type Obj1b
        Foo5b(obj=Obj1(slc=slice(3, 8)))

    print("Foo1", Foo1.parse_raw(foo1.json()).obj.__dict__)
    print("Foo2", Foo2.parse_raw(foo2.json()).obj.__dict__)
    print("Foo2b", Foo2b.parse_raw(foo2b.json()).obj.__dict__)
    print("Foo2c", Foo2c.parse_raw(foo2c.json()).obj.__dict__)
    print("Foo3", Foo3.parse_raw(foo3.json()).obj.__dict__)
    print("Foo4", Foo4.parse_raw(foo4.json()).obj.__dict__)
    print("Foo4b", Foo4b.parse_raw(foo4b.json()).obj.__dict__)
    print("Foo5", Foo5.parse_raw(foo5.json()).obj.__dict__)
    print("Foo5b", Foo5.parse_raw(foo5b.json()).obj.__dict__)

    # print(foo2.json())
    # print(Obj2.__validate__)

    # print("Bar1", Bar1.parse_raw(bar1.json()).obj.__dict__)

def test_plain_dataclass():
    """
    Test the special cases supporting serialization of plain dataclasses.
    These only support Serializable types and plain JSON types.
    """
    import json
    from scityping.json import scityping_encoder

    obj = DcStdTypes(i=1, s=("s","s"), f=3., slc=slice(1,2,3))

    data = {"i": 1,
            "s": ("s","s"),
            "f": 3.0,
            "slc": slice(1,2,3)}
    if KW_ONLY is None:  # Python ⩽ 3.9
        data["_"] = None
    assert Dataclass.reduce(obj) == ("scityping.base.Dataclass",
                                     (DcStdTypes, data))
    json_data = json.dumps(obj, default=scityping_encoder)
    if KW_ONLY is None:  # Python ⩽ 3.9
        assert json_data == '["scityping.base.Dataclass", [["scityping.base_types.Type", {"module": "test_dataclasses", "name": "DcStdTypes"}], {"i": 1, "s": ["s", "s"], "_": null, "f": 3.0, "slc": ["scityping.base_types.Slice", {"start": 1, "stop": 2, "step": 3}]}]]'
    else:  # Python ⩾ 3.10
        assert json_data == '["scityping.base.Dataclass", [["scityping.base_types.Type", {"module": "test_dataclasses", "name": "DcStdTypes"}], {"i": 1, "s": ["s", "s"], "f": 3.0, "slc": ["scityping.base_types.Slice", {"start": 1, "stop": 2, "step": 3}]}]]'
    objb = Dataclass.validate(json.loads(json_data))
    assert objb.slc == slice(1,2,3)

    # Test using Generic inside Union
    obj = DcStdTypes(i=1, s=("s","s"), f=[3., 1., 4.], slc=slice(1,2,3))
    json_data = json.dumps(obj, default=scityping_encoder)
    objb = Dataclass.validate(json.loads(json_data))
    assert objb.f == [3., 1., 4.]


# if __name__ == "__main__":
    # test_dataclass_and_pydantic()
    # test_plain_dataclass()


import pytest
from functools import partial
import operator
import numpy as np
from pydantic import ValidationError
import scityping
from scityping.base import Serializable
from scityping.utils import UnsafeDeserializationError
from scityping.pydantic import BaseModel
from scityping.functions import PureFunction, PartialPureFunction, CompositePureFunction

@PureFunction
def pure_f(x, n=2):
    """This is

    a multiline
    docstring
    """
    return x/n if x%n else x**n

def f(x, n=2):
    return x/n if x%n else x**n

class Foo(BaseModel):
    f: PureFunction

def test_purefunction():

    # Test reduction/reconstruction with Data object
    # Test that `config.trust_all_inputs` option works as expected
    data = Serializable.deep_reduce(pure_f)
    with pytest.raises(UnsafeDeserializationError):
        Serializable.validate(data)
    scityping.config.trust_all_inputs = True
    f2 = Serializable.validate(data)
    assert all(pure_f(x) == f2(x) for x in [3, 5, 8, 100, 101])

    # Nested PureFunctions get collapsed, so naturally they serialize fine
    fpp = PureFunction(PureFunction(pure_f))
    datapp = Serializable.deep_reduce(fpp)
    assert datapp == data
    fpp2 = Serializable.validate(datapp)
    assert all(fpp(x) == fpp2(x) for x in [3, 5, 8, 100, 101])

    # Test full round-trip serialization to JSON
    foo = Foo(f=pure_f)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

    # We want to force users to indicate that functions are pure;
    # therefore plain undecorated functions don't serialize.
    with pytest.raises(TypeError):
        Serializable.reduce(f)
    with pytest.raises(ValidationError):
        Foo(f=f)

    # We also support initializing from a string with the format
    f3 = PureFunction.validate("x -> x/(1+x)")
    foo = Foo(f=f3)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

def test_partialpurefunction():
    scityping.config.trust_all_inputs = True
    
    # If `partial` wraps a non-pure function, it cannot be serialized
    with pytest.raises(TypeError):
        Serializable.reduce(partial(f, n=3))

    # If `partial` wraps a *pure* function, it *can* be serialized
    # (Serializes as `PartialPure(f, kwds)`
    g = partial(pure_f, n=3)
    data = Serializable.reduce(g)
    g2 = Serializable.validate(data)
    assert all(g(x) == g2(x) for x in [3, 5, 8, 100, 101])

    # We can also directly declare a PartialPureFunction
    # (Serializes as `PartialPure(Pure, kwds)`. Note that this is different but equivalent to the case above.)
    h = PartialPureFunction(partial(f, n=4))
    data = Serializable.reduce(h)
    h2 = Serializable.validate(data)
    assert all(h(x) == h2(x) for x in [3, 5, 8, 100, 101])

    # Test full round-trip serialization to JSON
    foo = Foo(f=g)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

    foo = Foo(f=h)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

    # We also support initializing from a string with the format
    f3 = PureFunction.validate("x,y -> x/(1+x)")
    fa = partial(f3, 2)
    fb = partial(f3, y=2)

    foo = Foo(f=fa)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])
    foo = Foo(f=fb)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

def test_compositepurefunction():
    scityping.config.trust_all_inputs = True
    
    g = partial(pure_f, n=3)
    h = PartialPureFunction(partial(f, n=4))
    # Test arithmetic
    with pytest.raises(TypeError):  # At least for now, both arguments need to be PureFunctions
        w = pure_f + g      # (here g is just partial (although it wraps a PureFunction))
    w = pure_f + h          # (here h is a proper PureFunction)
    assert isinstance(w, CompositePureFunction)
    assert all(w(x) == f(x)+h(x) for x in [3, 5, 8, 100, 101])

    # Test reduction/reconstruction with Data object
    data = Serializable.reduce(w)
    w2 = Serializable.validate(data)
    assert all(w(x) == w2(x) for x in [3, 5, 8, 100, 101])

    # Test full round-trip serialization to JSON
    foo = Foo(f=w)
    foo2 = Foo.parse_raw(foo.json())
    assert all(foo.f(x) == foo2.f(x) for x in [3, 5, 8, 100, 101])

def test_special_functions():
    """
    Test built-in serialization support for common functions like
    NumPy ufuncs and array functions.
    """
    # ufunc: np.sin, np.add
    # array function: np.max, np.clip
    for fn in [operator.abs,
               np.sin,
               np.max,
               partial(np.add, 3),
               partial(np.clip, a_min=-1, a_max=2),
               partial(operator.pow, 2),
               "x -> np.tanh(x)"
               ]:
        purefn = PureFunction(fn)
        test_vals = [-10, -0.3, 0, 1, 5]
        # Test reduction/reconstruction with Data object
        data = Serializable.reduce(purefn)
        purefn2 = Serializable.validate(data)
        assert all(purefn(x) == purefn2(x) for x in test_vals)
        # Test full round-trip serialization to JSON
        foo = Foo(f=purefn)
        foo2 = Foo.parse_raw(foo.json())
        assert all(purefn(x) == foo.f(x) == foo2.f(x) for x in test_vals)




"""
Proxy module for Pydantic field types, which provides fallbacks in case
Pydantic is not installed.
"""

try:
    from pydantic import StrictStr, StrictBytes, StrictInt, StrictFloat, StrictBool
except ModuleNotFoundError:
    StrictStr = str
    StrictBytes = bytes
    StrictInt = int
    StrictFloat = float
    StrictBool = bool



"""
Tie-ins for pydantic which explicitely depend on pydantic types.
Provides:

- An scityping-aware "default" callable for `json.dump`, which allows it to
  serialize `Serializable` types.
- Drop-in replacements for `pydantic.BaseModel` and
  `pydantic.dataclasses.dataclass`, which by default uses the scityping JSON
  encoder provided by this package.
"""
from typing import TYPE_CHECKING, Any, Type
from functools import partial
from dataclasses import is_dataclass
from importlib.metadata import version
from pydantic import BaseModel as PydanticBaseModel
from pydantic.main import (ModelMetaclass as PydanticModelMetaclass,
                           ValidationError as PydanticValidationError)
from pydantic.generics import GenericModel as PydanticGenericModel
from pydantic.error_wrappers import display_errors
from pydantic.dataclasses import (dataclass as pydantic_dataclass,
                                  _validate_dataclass as _pydantic_validate_dataclass)
from .base import Serializable, json_like, Dataclass
from .json import scityping_encoder

if TYPE_CHECKING:
    from pydantinc.typing import ReprArgs

class ValidationError(PydanticValidationError):
    """
    This wraps ValidationError, and redefines the methods using `__name__`
    so they use `__qualname__` instead.
    """
    def __init__(self, orig_exception: PydanticValidationError):
        super().__init__(orig_exception.raw_errors, orig_exception.model)
        self._error_cache = orig_exception._error_cache

    def __str__(self) -> str:
        errors = self.errors()
        no_errors = len(errors)
        return (
            f'{no_errors} validation error{"" if no_errors == 1 else "s"} for {self.model.__qualname__}\n'
            f'{display_errors(errors)}'
        )

    def __repr_args__(self) -> 'ReprArgs':
        return [('model', self.model.__qualname__), ('errors', self.errors())]


def remove_overridden_validators(model: PydanticBaseModel) -> PydanticBaseModel:
    """
    Currently a Pydantic bug prevents subclasses from overriding root validators.
    (see https://github.com/samuelcolvin/pydantic/issues/1895)
    This function inspects a Pydantic model and removes overriden
    root validators based of their `__name__`.
    Assumes that the latest entries in `__pre_root_validators__` and
    `__post_root_validators__` are earliest in the MRO, which seems to be
    the case.
    """
    model.__pre_root_validators__ = list(
        {validator.__name__: validator
         for validator in model.__pre_root_validators__
         }.values())
    model.__post_root_validators__ = list(
        {validator.__name__: (skip_on_failure, validator)
         for skip_on_failure, validator in model.__post_root_validators__
         }.values())
    return model

class ModelMetaclass(PydanticModelMetaclass):
    """
    Use this as the metaclass for BaseModel types, to allow them

    1. to find serializers defined in Serializable;
    2. to allow their subclasses to override root validators,
       (patch for https://github.com/samuelcolvin/pydantic/issues/1895)
       (Only applied if Pydantic version is ⩽ 1.8)
    """
    def __new__(mcs, name, bases, namespace, **kwargs):
        ann = namespace.get("__annotations__", {})
        for field, T in ann.items():
            if is_dataclass(T) and not isinstance(T, Serializable):  # Serializable subclasses already have an __get_validators which yields cls.validate
                orig_get_validators = getattr(T, "__get_validators__", None)
                if not orig_get_validators:
                    try:
                        T.__get_validators__ = _get_dataclass_validator(T)
                    except AttributeError:
                        logger.error(
                            "Cannot add attribute '__get_validators__' to the "
                            f"type `{T.__qualname__}`: the class `{name}` will "
                            "not be deserializable. This can happen if the class "
                            "was defined with '__slots__'.")
        obj = super().__new__(mcs, name, bases, namespace, **kwargs)
        # Wrap the reduce with scityping_encoder
        encoder = partial(scityping_encoder, base_encoder=obj.__json_encoder__)
        obj.__json_encoder__ = staticmethod(encoder)
        # If Pydantic version is ⩽ 1.8, apply patch to allow their subclasses to override root validators.
        _pydantic_version = tuple(int(i) for i in version("pydantic").split("."))
        if _pydantic_version <= (1, 8):
            obj = remove_overridden_validators(obj)
        return obj

class BaseModel(PydanticBaseModel, metaclass=ModelMetaclass):
    pass

class GenericModel(PydanticGenericModel, BaseModel):
    pass

def _get_dataclass_validator(T: type):
    def _dataclass_type_match(value):
        if not isinstance(value, T):
            raise TypeError(f"Value is not of type `{T.__qualname__}`")
        return value
    def _get_validators():
        yield _dataclass_deserializer
        yield _dataclass_type_match
    return _get_validators

def _dataclass_deserializer(value):
    if json_like(value, Dataclass._registry):
        return Dataclass.validate(value)
    else:
        return value

# C.f. pydantic.dataclasses
def _validate_serializable_dataclass(cls: Type["DataclassT"], v: Any) -> "DataclassT":
    if json_like(v, cls._registry):
        return cls.validate(v)
    else:
        return _pydantic_validate_dataclass(cls, v)

# Hard-coded special-case support for stdlib dataclasses
def _validate_plain_dataclass(cls: Type["DataclassT"], v: Any) -> "DataclassT":
    if json_like(v, Dataclass._registry):
        return Dataclass.validate(v)
    else:
        return _pydantic_validate_dataclass(cls, v)

def dataclass(_cls=None, **kwargs):
    ## Secondary code path: decorator with arguments ##
    if _cls is None:
        return partial(dataclass, **kwargs)

    ## Primary code path: decorator without arguments ##

    # Validation in a Pydantic dataclass is performed by storing a BaseModel
    # in the private attribute __pydantic_model__. Additional dunder methods are
    # added to the dataclass: The validation logic for the class itself (i.e.
    # what is used when the class is part of a BaseModel) is attached to __validate__.
    # First we create the dataclass, then patch the validation & serialization
    # logic so they recognize Serializable types
    dclass = pydantic_dataclass(_cls, **kwargs)

    # Validation: Patch __validate__ to recogize a Serialized `self`
    #    Note that any Serializable *arguments* to the dataclass will already
    #    be correctly deserialized. What we need to catch is the case where
    #    `self` is a dataclass, and we are instantiating it from serialized data.
    #    Therefore we only need to wrap __validate__, not __pydantic_validate_values__
    if issubclass(dclass, Serializable):
        setattr(dclass, "__validate__", classmethod(_validate_serializable_dataclass))
    else:
        # Hard-coded special-case support for stdlib dataclasses
        setattr(dclass, "__validate__", classmethod(_validate_plain_dataclass))
    
    # JSON encoder: As in `ModelMetaclass`, we patch the __json_encoder__ attribute
    obj = dclass.__pydantic_model__
    encoder = partial(scityping_encoder, base_encoder=obj.__json_encoder__)
    obj.__json_encoder__ = staticmethod(encoder)

    # Return
    return dclass



import numpy as np
import torch
from scityping.torch import TorchTensor, TorchModule, TorchGenerator
from scityping.torch import torch_module_state_encoder, torch_module_state_decoder

def test_torch():
    # Scalars
    s_u8   = torch.tensor(np.uint8(5))
    s_i32  = torch.tensor(np.int32(5))
    s_f64  = torch.tensor(np.float64(5))
    s_bool = torch.tensor(True)
    # Vectors
    v_u8   = torch.tensor(np.uint8  ([5, 8, 9]))
    v_i32  = torch.tensor(np.int32  ([5, 8, 9]))
    v_f64  = torch.tensor(np.float64([5, 8, 9]))
    v_bool = torch.tensor([True, False, True])
    # Matrices
    m_u8   = torch.outer(v_u8, v_u8)
    m_i32  = torch.outer(v_i32, v_i32)
    m_f64  = torch.outer(v_f64, v_f64)
    m_bool = torch.outer(v_bool, v_bool)

    for a in [s_u8, s_i32, s_f64, s_bool,
              v_u8, v_i32, v_f64, v_bool,
              m_u8, m_i32, m_f64, m_bool]:
        b = TorchTensor.validate(TorchTensor.reduce(a))
        assert a.shape == b.shape
        assert a.dtype == b.dtype
        assert (a == b).all()


    # TODO: Test TorchModule, TorchGenerator
    #       Test TorchTensor & TorchGenerator with GPU tensors, if GPU is available



# -*- coding: utf-8 -*-
"""
A set of Pydantic-compatible types relevant for scientific applications.
Includes types for numbers, NumPy, and PyTorch subtypes and objects.

"""
# TODO: __modify_schema__ should be defined in Serializable, by using the schema of the Data

from __future__ import annotations

import abc
import typing
import inspect
from types import SimpleNamespace
from collections.abc import Callable as Callable_
from typing import Optional, Union, Any, Iterable, List, Tuple, Dict, Literal
from dataclasses import dataclass, asdict, fields

import numbers

import logging
logger = logging.getLogger(__name__)

from .base import Serializable, json_like, Dataclass, validate_dataclass
from .utils import get_type_key
from .config import config

# dataclasses are used for the `Data` container associated to each type.
# # Use Pydantic dataclasses if available (they provide serialization/deserialization)
# # Use the smart_union=True option with Pydantic dataclasses, to avoid accidental coercion
# # Otherwise, use builtin dataclasses
# # TODO: Don’t use pydantic dataclasses. Use normal ones, and include deserialization in `SerializedData`
# try:
#     import pydantic
# except ModuleNotFoundError:
#     from dataclasses import dataclass
# else:
#     from .pydantic import dataclass as pydantic_dataclass
#     dataclass = pydantic_dataclass(config={"smart_union": True})

__all__ = ["dataclass", "SerializedData",
           "NoneType", "Complex", "Range", "Slice",
           "Number", "Integral", "Real",
           "Type", "GenericType", "PydanticGenericType"]

# ###############
# Helper base type for the nested `Data` classes
# This is used to reduce boilerplate (remove the need for @dataclass and `self` in arguments)

class SerializedDataMeta(abc.ABCMeta):
    def __new__(metacls, cls, bases, namespace):
        for nm, attr in namespace.items():
            if not nm.startswith("__") and isinstance(attr, Callable_):
                # NB: staticmethod and classmethod objects are not callable
                # Dunder methods should remain methods
                sig = inspect.signature(attr)
                if sig.parameters and next(iter(sig.parameters)) == "self":
                    # If the first argument is `self`, don’t convert to a static method
                    pass
                else:
                    namespace[nm] = staticmethod(attr)
        obj = super().__new__(metacls, cls, bases, namespace)
        return dataclass(obj)
class SerializedData(Dataclass, metaclass=SerializedDataMeta):
    """
    Define a helper base class for nested Data data classes, which
    - makes every method static (avoiding the need to pass an unnecessary 'self');
    - applies the @dataclass decorator to the new class, converting it to a dataclass.
    """
    @abc.abstractmethod
    def encode(value): raise NotImplementedError
    def __post_init__(self):
        validate_dataclass(self, inplace=True)
    # ## Pydantic compatibility ##
    # # Note that instead of defining these methods, `SerializedData` could inherit
    # # from `BaseModel`, but that would make pydantic a hard dependency
    # @classmethod
    # def __get_validators__(cls):
    #     yield cls.validate
    # @classmethod
    # def validate(cls, value, field=None):  # 'field' only present for consistency
    #     return cls(**value)

# ###############
# Special type which only accepts the value `None`; used for deactivating
# fields in a subclass

class NoneType:
    """
    A Pydantic compatible 'type' which only accepts None.
    Use case: deactivating options of a parent class by forcing them to be None
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, v, field=None):
        if field is None:
            field = SimpleNamespace(name="")
        assert v is None, f"Field '{field.name}' of '{cls.__qualname__}' accepts only the value `None`."
        return None

# ###############
# Support for builtin types:
# - complex
# - range
# - slice
# - type

# TODO: Since they only contain plain data types, we could make the `Data` classes more lightweight by just subclassing `tuple`.

class Complex(Serializable, complex):  # Using same name allows serializing original type
  class Data(SerializedData):
    real: float
    imag: float
    def encode(z): return z.real, z.imag
    def decode(data): return complex(data.real, data.imag)

class Range(Serializable):  # NB: It is not permitted to subclass `range`
    class Data(SerializedData):
        start: int
        stop: Optional[int]=None
        step: Optional[int]=None
        def encode(r): return r.start, r.stop, r.step
        def decode(data): return range(data.start, data.stop, data.step)  # Required since we don’t subclass `range`

    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     """We need to tell pydantic how to export this type to the schema,
    #     since it doesn't know what to map the type to.
    #     """
    #     # See https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types
    #     # for the API, and https://pydantic-docs.helpmanual.io/usage/schema/
    #     # for the expected fields
    #     # TODO: Use 'items' ?
    #     field_schema.update(
    #         type="array",
    #         description="('range', START, STOP, STEP)"
    #         )
Range.register(range)

class Slice(Serializable):  # NB: It is not permitted to subclass `range`
    class Data(SerializedData):
        start: int
        stop: Optional[int]=None
        step: Optional[int]=None
        def encode(r): return r.start, r.stop, r.step
        def decode(data): return slice(data.start, data.stop, data.step)  # Required since we don’t subclass `slice`

    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     """We need to tell pydantic how to export this type to the schema,
    #     since it doesn't know what to map the type to.
    #     """
    #     # See https://pydantic-docs.helpmanual.io/usage/types/#custom-data-types
    #     # for the API, and https://pydantic-docs.helpmanual.io/usage/schema/
    #     # for the expected fields
    #     # TODO: Use 'items' ?
    #     field_schema.update(
    #         type="array",
    #         description="('slice', START, STOP, [STEP])"
    #         )
Slice.register(slice)

# ###############
# Overrides of typing types

#  Pydantic recognizes typing.Sequence, but treats it as a shorthand for
#  Union[List, Tuple] (but with equal precedence). This means that other
#  sequence types like `range` are not recognized.
Sequence = Union[Range, typing.Sequence]

# ###############
# Types based on numbers module

class Number(numbers.Number):
    """
    This type does not support coercion, only validation.
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, value, field=None):
        if field is None:
            field = SimpleNamespace(name="")
        if not isinstance(value, numbers.Number):
            raise TypeError(f"Field {field.name} expects a number. "
                            f"It received {value} [type: {type(value)}].")
        return value
    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type="number")

class Integral(numbers.Integral):
    """
    This type does not support coercion, only validation.
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, value, field=None):
        if field is None:
            field = SimpleNamespace(name="")
        if not isinstance(value, numbers.Integral):
            raise TypeError(f"Field {field.name} expects an integer. "
                            f"It received {value} [type: {type(value)}].")
        return value
    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type="integer")

class Real(numbers.Real):
    """
    This type generally does not support coercion, only validation.
    Exception: integer values are cast with `float`.
    """
    @classmethod
    def __get_validators__(cls):
        yield cls.validate
    @classmethod
    def validate(cls, value, field=None):
        if field is None:
            field = SimpleNamespace(name="")
        if not isinstance(value, numbers.Real):
            raise TypeError(f"Field {field.name} expects a real number. "
                            f"It received {value} [type: {type(value)}].")
        elif isinstance(value, numbers.Integral):
            # Convert ints to floating point
            return float(value)
        return value
    # # TODO
    # @classmethod
    # def __modify_schema__(cls, field_schema):
    #     field_schema.update(type="number")


#####################
# Type
#####################

T = typing.TypeVar('T')
class Type(typing.Type[T], Serializable):  # NB: Serializable must come 2nd
    """
    Make types serializable; the serialization format is
        ('Type', <module name>, <type name>)
    During deserialization, it effectively executes
        from <module name> import <type name>

    .. Caution:: **Limitation** As with `typing.Type`, one can indicate the specific type between
       brackets; e.g. ``Type[int]``, and Pydantic will enforce this restriction.
       However at present deserialization only works when the type is unspecified.

    .. Warning:: This kind of serialization will never be 100% robust and
       should be used with care. In particular, since it relies on <module name>
       remaining unchanged, it is certainly not secure. (Although no less so
       than `pickle`.)
       Because of the potential security issue, it requires adding modules where
       tasks are defined to the ``scityping.config.safe_packages`` whitelist.
    """
    # FIXME: When the type T is specified, the specialized type doesn't inherit __get_validators__
    #   (although it does inherit the other methods)
    class Data(SerializedData):
        module: str
        name: str

        @classmethod
        def encode(cls, T: typing.type) -> Type.Data:
            if not isinstance(T, type):
                raise TypeError(f"'{T}' is not a type.")
            if T.__module__ == "__main__":
                # raise ValueError("Can't serialize types defined in the '__main__' module.")
                logger.warning(f"Type '{T}' was serialized in the '__main__' module; "
                               "it will only be deserializable within the same script "
                               "or notebook.")
            return cls(T.__module__, T.__qualname__)

        def decode(data: Type.Data) -> Type:
            from importlib import import_module
            module = data.module
            if (any(module.startswith(sm) for sm in config.safe_packages)
                  or config.trust_all_inputs):
                m = import_module(module)
                for _name in data.name.split("."):
                    try:
                        m = getattr(m, _name)
                    except AttributeError as e:
                        if module == "__main__":
                            raise RuntimeError(
                                f"The type '{data.name}' was serialized in the same "
                                "namespace in which it was defined; therefore it can "
                                "also only be deserialized in that namespace. "
                                "However no type of that name was found in the current "
                                "`__main__` namespace")
                        else:
                            raise e
                else:  # All dotted name elements were found
                    T = m
            else:
                raise RuntimeError(
                    "As with pickle, deserialization of types can lead to "
                    "arbitrary code execution. It is only permitted after "
                    f"adding '{module}' to ``scitying.config.safe_packages`` "
                    "(recommended) or setting the option "
                    "``scityping.config.trust_all_inputs = True``.")
            return T
            
    @classmethod
    def __modify_schema__(cls, field_schema):
        field_schema.update(type='array',
                            items=[{'type': 'string'}]*3)


class GenericType(Type):
    class Data(SerializedData):
        baseT: Type
        args: List[Type]

        def encode(T: typing._GenericAlias) -> Union[GenericType.Data,Type.Data]:
            if T.__parameters__:  # __parameters__ is the list of non specified types
                if not all(isinstance(argT, typing.TypeVar) for argT in T.__args__):
                    raise NotImplementedError(
                        "We only support generic types for which either non of the "
                        "type arguments are specified, or all of them are.\n"
                        f"Type {T} has both.")
                # For non-concrete types, the standard Type serialization format suffices
                # (NB: We can't use Type.reduce, because we need '_name' instead of '__name__'
                # FIXME: I don't think returning a Data of another type is supported yet
                return Type.Data(module=T.__module__, name=T._name)
                # raise NotImplementedError("Only concrete generic types can be serialized. "
                #                           "(So e.g. `List[int]`, but not `List`.)")
            if T.__module__ == "__main__":
                raise ValueError("Can't serialize types defined in the '__main__' module.")
            return T, T.__args__

        def decode(data: GenericType.Data) -> GenericType:
            return data.baseT[tuple(data.args)]

class PydanticGenericType(GenericType):
    class Data(GenericType.Data):
        def encode(T: pydantic.main.ModelMetaclass) -> PydanticGenericType.Data:
            # Get the base Generic type and type arguments
            #   E.g. For `Foo[int]`, retrieve `Foo` and `(int,)`
            # NB: In contrast to normal generic types, Pydantic Generics don't have
            #   __origin__ or __args__ attributes. But Pydantic maintains a cache
            #   of instantiated generics, to avoid re-instantiating them; this
            #   cache is keyed by the base generic type, and the argument types
            try:
                genT, paramT = next(k for k, v in pydantic.generics._generic_types_cache.items()
                                    if v is T)
            except StopIteration:
                # The cache is updated as soon as a concrete generic type is first created,
                # i.e. the first time `Foo[int]` appears.
                # The only way it should happen that T is not in the cache, is if
                # T is a pure generic type. In this case, the normal Type serializer
                # works fine.
                return Type.Data.encode(T)
            
            return (genT, paramT)



import pickle
# import numpy as np
# import scipy.stats as stats
# import scityping.numpy as stnp
# import scityping.scipy as stsp
from scityping.functions import PureFunction
from scityping.config import config
from functools import partial

def test_pickling():

    old_config = dict(trust_all_inputs=config.trust_all_inputs)
    config.trust_all_inputs = True  # Allow deserializing functions

    @PureFunction
    def f(x, n):
        return x**n

    # NB: If we didn't support pickling, this would fail because f() is
    # defined in the local scope.
    fbytes = pickle.dumps(f)
    f2 = pickle.loads(fbytes)

    assert f2 is not f
    assert f2(4, 6) == f(4, 6)

    g = partial(f, n=3)

    gbytes = pickle.dumps(g)
    g2 = pickle.loads(gbytes)

    assert g2 is not g
    assert g2(6) == f(6, n=3)

    config.trust_all_inputs = old_config["trust_all_inputs"]



"""
Pydantic-aware classes for Quantities objects.
"""

import numpy as np
import quantities as pq

from .base import Serializable, ABCSerializable
from .base_types import SerializedData
from .numpy import Array

class QuantitiesQuantity(Serializable, pq.Quantity):
    class Data(SerializedData):
        data : Array  # Names match signature of pq.Quantity
        units: str
        def encode(val): return v.magnitude, str(v.dimensionality)

Quantity = QuantitiesQuantity  # Match the name in quantities

type_to_register = pq.Quantity
cls = QuantitiesQuantity
ABCSerializable.register(type_to_register)
ABCSerializable._base_types[cls] = type_to_register
for C in cls.mro():
    if issubclass(C, Serializable):
        C._registry[cls] = cls
        C._registry[type_to_register] = cls


class QuantitiesUnit(QuantitiesQuantity):
    """
    Exactly the same as QuantitiesValue, except that we enforce the magnitude
    to be 1. In contrast to Pint, Quantities doesn't seem to have a unit type.
    """
    class Data(QuantitiesQuantity.Data):
        pass
    @classmethod
    def validate(cls, v, field=None):
        v = super().__init__(v, field=field)
        if v.magnitude != 1:
            field_name = field.name if field else "<QuantitiesUnit>"
            raise ValueError(f"Field {field_name}: Units must have "
                             "magnitude of one.")
        return v

class QuantitiesDimension(Serializable, pq.dimensionality.Dimensionality):
    class Data(SerializedData):
        name: str
        def encode(dim): return str(dim)
        def decode(data): return pq.quantity.validate_dimensionality(data.name)

    # Allow casting from Quantity
    @classmethod
    def validate(cls, v, field=None):
        if isinstance(v, pq.Quantity):
            v = v.dimensionality
        return super().validate(v, field=field)



import logging
import dataclasses
import pytest

import numpy as np
import pint
from pydantic import validator
from pydantic import ValidationError
from scityping.pydantic import BaseModel, dataclass

from typing import List, Tuple
from scityping.numpy import NPValue, Array, RNGenerator, RandomState, SeedSequence

# TODO: Systematically test casting of NPValue and Array, especially with
#       generic types
# TODO: Test Torch types

def test_numpy(caplog):

    arr = np.arange(3, dtype='uint32')
    arr2 = Array.validate(Array.reduce(arr))
    assert repr(arr) == repr(arr2)  # Check that they match exactly, including dtype

    class Model(BaseModel):
        a: float
        # dt: float=0.01
        dt: float
        def integrate(self, x0, T, y0=0):
            x = x0; y=y0
            a = self.a; dt = self.dt
            for t in np.arange(0, T, self.dt):
                x += y*dt; y+=a*x*dt
            return x
        @property
        def params(self):
            return ParameterSet(self.dict())

    m = Model(a=-0.5, dt=0.01)
    assert repr(m) == "Model(a=-0.5, dt=0.01)"
    assert np.isclose(m.integrate(1, 5.), -0.924757713688715)
    assert type(m.a) is float

    class Model2(Model):
        b: NPValue[np.float64]
        w: Array[np.float32]
        def integrate(self, x0, T, y0=0):
            x = w*x0; y = w*y0
            α = np.array([self.a, self.b]); dt = self.dt
            for t in np.arange(0, T, self.dt):
                x += y*dt; y+=α*x*dt
            return x.sum()

    # Test subclassed List and Tuple
    # TODO: tuples with floats
    class ModelB(BaseModel):
        l1: list
        l2: List[int]
        t1: tuple
        # t2: Tuple[np.float16, np.float32]
        pass

    mb = ModelB(l1=['foo', 33], l2=[3.5, 3], t1=(1, 2, 3, 4, 5))
    assert mb.l1 == ['foo', 33]
    assert mb.l2 == [3, 3]
    assert mb.t1 == (1, 2, 3, 4, 5)

    # NOTE: Some of the tests below do not work presently with
    #       dataclasses, due to the way keyword vs non-keyword parameters
    #       are ordered (see discussion https://bugs.python.org/issue36077)
    #       Pydantic models however allows setting default/computed
    #       override base class attributes, without having to use a default for
    #       new every attribute in the derived class
    class Model3(Model):
        a: NPValue[float]  = 0.3
        β: NPValue[float] = None
        @validator('β', pre=True, always=True)
        def set_β(cls, v, values):
            a, dt = (values.get(x, None) for x in ('a', 'dt'))
            # This test is important, because if `a` or `dt` are missing
            # (as when we assign to `m4_fail` below) or trigger a validation
            # error, they will be missing from `values` and raise KeyError here,
            # (See https://pydantic-docs.helpmanual.io/usage/validators)
            if all((a, dt)):
                return a*dt
        def integrate(self, x0, T, y0=0):
            x = x0; y=y0
            β = self.β; dt = self.dt
            for t in np.arange(0, T, self.dt):
                x += y*dt; y+=β*x
            return x

    # NOTE: Following previous note: with vanilla dataclasses Model4 would need
    #       to define defaults for every attribute.
    class Model4(Model3):
        b: NPValue[np.float32]
        w: Array[np.float32] = (np.float32(1), np.float32(0.2))
        γ: Array[np.float32] = None
        @validator('γ', pre=True, always=True)
        def set_γ(cls, v, values):
            a, b, β = (values.get(x, None) for x in ('a', 'b', 'β'))
            if all((a, b, β)):
                return β * np.array([1, b/a], dtype='float32')
        def integrate(self, x0, T, y0=0):
            x = w*x0; y = w*y0
            γ = self.γ; dt = self.dt
            for t in np.arange(0, T, self.dt):
                x += y*dt; y+=γ*x
            return x.sum()

    # Basic type conversion
    w64 = np.array([0.25, 0.75], dtype='float64')
    w32 = w64.astype('float32')
    w16 = w64.astype('float16')
    with pytest.raises(ValidationError):   # With vanilla dataclass: TypeError
        largeb = 57389780668102097303  # int(''.join((str(i) for i in np.random.randint(10, size=20))))
        m2_np = Model2(a=-0.5, b=largeb, w=w16, dt=0.01)
    m2_np = Model2(a=-0.5, b=-0.1, w=w16, dt=0.01)
    assert repr(m2_np) == "Model2(a=-0.5, dt=0.01, b=-0.1, w=array([0.25, 0.75], dtype=float32))"
    assert m2_np.b.dtype is np.dtype('float64')
    # Test conversion from ints and list
    m2_nplist = Model2(a=-0.5, b=np.float32(-0.1), w=[np.float32(0.25), np.float32(0.75)], dt=0.01)
    assert m2_nplist.w.dtype is np.dtype('float32')

    # Test computed fields and numpy type conversions
    m3 = Model3(a=.25, dt=0.02)
    assert type(m3.a) != type(m.a)
    assert type(m3.a) is np.dtype(float).type
    assert type(m3.β) is np.dtype(float).type
    assert m3.β == 0.005
    # These two constructors seem to specify the same model (default w is
    # simply reproduced), but because validators are not called on default
    # arguments in the first case the default w remains a tuple.
    # You can use the following validator to force its execution
    # @validator('w', always=True)
    #     def set_w(cls, v):
    #         return v
    m4_default = Model4(a=.25, dt=0.02, b=np.float32(0.7))
    m4         = Model4(a=.25, dt=0.02, b=np.float32(0.7),
                        w=(np.float32(1), np.float32(0.2)))
    assert isinstance(m4_default.w, tuple)
    assert isinstance(m4.w, np.ndarray) and m4.w.dtype.type is np.float32

    m4_16 = Model4(a=.25, dt=0.02, b=np.float32(0.3), w=w16)
    m4_32 = Model4(a=.25, dt=0.02, b=np.float32(0.3), w=w32)
    assert m4_16.w.dtype == 'float32'
    assert m4_32.w.dtype == 'float32'
    # Casts which raise TypeError
    with pytest.raises(ValidationError):   # With stdlib dataclass: TypeError
        m4_fail = Model4(a=.3, b=0.02, w=w64)
    with pytest.raises(ValidationError):   # With stdlib dataclass: TypeError
        wint = np.array([1, 2])
        m2_npint = Model2(a=-0.5, b=-0.1, w=wint, dt=0.01)

    # Import after defining Model to test difference in mapping of `float` type

def test_seedseq():
    class SeedModel(BaseModel):
        seedseq: SeedSequence

    from numpy.random import SeedSequence as SeedSeq

    seedseq1 = SeedSeq()
    seedseq2 = SeedSeq(452)
    seedseq3 = SeedSeq(spawn_key=(2,3))
    seedseq4 = SeedSeq((212,333), spawn_key=(2,3))
    seedseq5 = SeedSeq(212, spawn_key=(2,3), pool_size=8)

    seedseq3.spawn(2)  # Increments n_children_spawned
    seedseq5.spawn(3)  # Increments n_children_spawned

    for seedseq in [seedseq1, seedseq2, seedseq3, seedseq4, seedseq5]:
        seedseqb = SeedModel.parse_raw(SeedModel(seedseq=seedseq).json()).seedseq
        assert seedseq.generate_state(1)[0] == seedseq.generate_state(1)[0]
        # If n_children_spawned was serialized, spawning new children will start
        # where the pre-serialized SeedSequence left off
        child = seedseq.spawn(10)[7]
        childb = seedseqb.spawn(10)[7]
        assert child.generate_state(1)[0] == childb.generate_state(1)[0]

def test_rng():
    class RandomModel(BaseModel):
        rng: RNGenerator

    # # FIXME
    # RandomModel.schema()

    from numpy.random import Generator, PCG64, MT19937, SFC64, Philox
    seed = 953235987

    rm_pcg = RandomModel(rng=Generator(PCG64(seed)))
    rm_mt  = RandomModel(rng=Generator(MT19937(seed)))
    rm_sfc = RandomModel(rng=Generator(SFC64(seed)))
    rm_phi = RandomModel(rng=Generator(Philox(seed)))

    # Save models in their current initialized state
    pcg_json = rm_pcg.json()
    mt_json  = rm_mt.json()
    sfc_json = rm_sfc.json()
    phi_json = rm_phi.json()

    # Draw from models, advancing the bit generator
    pcg_draws = rm_pcg.rng.random(size=5)
    mt_draws  = rm_mt.rng.random(size=5)
    sfc_draws = rm_sfc.rng.random(size=5)
    phi_draws = rm_phi.rng.random(size=5)

    # Create new model copies, in their original initialized states
    rm_pcg2 = RandomModel.parse_raw(pcg_json)
    rm_mt2  = RandomModel.parse_raw(mt_json)
    rm_sfc2 = RandomModel.parse_raw(sfc_json)
    rm_phi2 = RandomModel.parse_raw(phi_json)

    # Draw again => same numbers as before
    assert np.all(pcg_draws == rm_pcg2.rng.random(size=5))
    assert np.all(mt_draws  == rm_mt2.rng.random(size=5))
    assert np.all(sfc_draws == rm_sfc2.rng.random(size=5))
    assert np.all(phi_draws == rm_phi2.rng.random(size=5))

    # Drawing _again_ produces different numbers => these really are random numbers
    assert np.all(pcg_draws != rm_pcg2.rng.random(size=5))
    assert np.all(mt_draws  != rm_mt2.rng.random(size=5))
    assert np.all(sfc_draws != rm_sfc2.rng.random(size=5))
    assert np.all(phi_draws != rm_phi2.rng.random(size=5))

def test_legacy_rng():
    class RandomModel(BaseModel):
        rng: RandomState

    # # FIXME
    # RandomModel.schema()

    seed = 953235987
    rm_leg = RandomModel(rng=RandomState(seed))

    # Save models in their current initialized state
    leg_json = rm_leg.json()

    # Draw from models, advancing the bit generator
    leg_draws = rm_leg.rng.random(size=5)

    # Create new model copies, in their original initialized states
    rm_leg2 = RandomModel.parse_raw(leg_json)

    # Draw again => same numbers as before
    assert np.all(leg_draws == rm_leg2.rng.random(size=5))

    # Drawing _again_ produces different numbers => these really are random numbers
    assert np.all(leg_draws != rm_leg2.rng.random(size=5))






import sys
import re
import logging
from collections import defaultdict
from itertools import tee

logger = logging.getLogger(__name__)

class NotFound:
    pass

class UnsafeDeserializationError(RuntimeError):
    pass

def deref_module_name(val):
    """
    If `val` is a string, assume it is an import path to that module and
    return that module.
    Otherwise, return `val`.
    """
    if "." not in val:
        raise NotImplementedError(
            "I'm not sure in which situation one would need a str value "
            "that doesn't include a module; waiting for a use case.")
    else:
        # 'val' is a dotted name: first part indicates the package
        # where type is defined.
        # Problem: given "foo.bar.ObjName", we don't know whether to do
        # `from foo.bar import ObjName` or `from foo import bar.ObjName`.
        # So we try all combinations, giving precedence to more
        # specified packages (so `from foo.bar` would be tried first).
        # For each combination, we first check whether that package is
        # imported:  if not, it is not possible (without irresponsible
        # manipulation of imports) for that path to `ObjName` to be
        # defined.
        modname = val
        name = ""
        for k in range(val.count(".")):
            modname, name_prefix = val.rsplit(".", 1)
            name = name_prefix + ("." + name if name else "")
            mod = sys.modules.get(modname)
            if mod:  # Only attempt to import from already loaded modules
                try:
                    val = getattr(mod, name)
                except AttributeError:
                    pass  # Continue trying with a less specified import path
                else:
                    break
        else:
            # If we exit the loop without finding anything, it means the
            # required module is not available.
            raise ModuleNotFoundError(f"The value {val} does not match "
                                      "any imported module.")

    return val

class ModuleList(list):
    def __iter__(self):
        """
        Return elements from the list. Strings are replaced by importing the
        corresponding module.
        """
        baseiter = super().__iter__()
        for i, el in enumerate(baseiter):
            if isinstance(el, str):
                el = deref_module_name(el)
                # Update list so module is not imported again
                self[i] = el
            yield el

class LazyDict(dict):
    """A dictionary allowing to specify objects that haven’t been loaded yet.

    A string value is interpreted as the import path to an object. This is
    useful for specifying types without requiring them to be imported, or which
    may not always be available.

    .. Caution:: Should not be used to store `str` values which are not imports.
    """
    def __getitem__(self, key):
        val = super().__getitem__(key)
        if isinstance(val, str):
            newval = deref_module_name(val)
            # Store value so it doesn’t need to be imported again
            self[key] = newval
            val = newval
        return val

    def get(self, key, default=None):
        val = super().get(key, default)
        if val != default and isinstance(val, str):  # I think it’s less surprising if the default is always unmodified ?
            try:
                newval = deref_module_name(val)
            except ModuleNotFoundError:
                return default  # EARLY EXIT
            # Store value so it doesn’t need to be imported again
            self[key] = newval
            val = newval
        return val


def get_type_key(obj: type):
    try:
        serialization_key = obj.__qualname__
        module = obj.__module__
    except AttributeError:
        raise TypeError("`get_type_key` accepts only instances of `type`.")
    # Allow types to override their serialization key; this is used by generics like PureFunction
    serialization_key = getattr(obj, "__serialization_key__", serialization_key)
    return f"{module}.{serialization_key}"

class TypeRegistry(dict):
    """A registry mapping serializable type keys to the types themselves.

    A `TypeRegistry` attached to serializable type A is used to store the list
    of serialiable types which are compatible with A (loosely speaking, values
    of the registry are subtypes of A).
    To allow fuzzy matching, we associate each subtype with a string key which
    is all lowercased and can be split into tokens separated by a period ".".
    Keys are computed deterministically from a type, and there can be multiple
    keys pointing to the same serializable type; this allows a single
    serializer to be used for multiple types.
    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._key_tokens = {}
        for k, v in dict(*args, **kwargs).items():
            self[k] = v  # Use __setitem__ which has the key standardization

    def __contains__(self, key):
        """Implement case-insensitive key matching."""
        if isinstance(key, type):
            key = get_type_key(key)
        elif not isinstance(key, str):
            raise TypeError("`TypeRegistry` accepts only types or strings "
                            f"as keys; received: {key}")
        return super().__contains__(key.lower())

    def __setitem__(self, key, value):
        if isinstance(key, type):
            key = get_type_key(key)
        elif not isinstance(key, str):
            raise TypeError("`TypeRegistry` accepts only types or strings "
                            f"as keys; received: {key}")
        key = key.lower()
        tokens = frozenset(tok for tok in key.split("."))
        if tokens in self._key_tokens:
            if self._key_tokens[tokens] == key and self.get(key.lower()) is value:
                # Some solutions for special cases may be  easier if we allow
                # setting registry values more than once. As long as the value
                # is the same, this is fine.
                return  # EARLY EXIT
            else:
                logger.error("The Serializable TypeRegistry already contained an "
                             f"entry for key {self._key_tokens[tokens]}, which "
                             "has been replaced.")
                super().__delitem__(self._key_tokens[tokens])
        super().__setitem__(key, value)
        self._key_tokens[tokens] = key

    def __delitem__(self, key):
        if isinstance(key, type):
            key = get_type_key(key)
        elif not isinstance(key, str):
            raise TypeError("`TypeRegistry` accepts only types or strings "
                            f"as keys; received: {key}")
        key = key.lower()
        tokens = frozenset(tok for tok in key.split("."))
        super().__delitem__(key)
        del self._key_tokens[tokens]

    def __getitem__(self, key):
        if isinstance(key, type):
            key = get_type_key(key)
        elif not isinstance(key, str):
            raise TypeError("`TypeRegistry` accepts only types or strings "
                            f"as keys; received: {key}")
        key = key.lower()

        # Fast initial search: use subset matching to find a) the largest number
        # of tokens which match some of the keys, and b) the corresponding keys
        ordered_tokens = tuple(tok for tok in  key.split("."))
        tokens = set(ordered_tokens)
        matches = defaultdict(lambda: [])
        for regtokens, regkey in self._key_tokens.items():
            if regtokens & tokens:
                matches[len(regtokens & tokens)].append((regtokens, regkey))

        # Slower filtering based on matching RTL
        if not matches:
            raise KeyError(f"Registry contains no entry matching {key}")
        else:
            matches = matches[max(matches)]  # Look only at keys which match the most tokens
            pattern = r"(?:[a-zA-Z0-9_\-.[]()]+\.|)"  # Pattern allows for missing modules between tokens
            rtl_matches = [regkey for (regtokens, regkey) in matches if
                           re.search(pattern  # NB: We prepend 'pattern' but don't postpend it (the actual type name should match)
                                     + pattern.join(tok.replace("[",r"\[").replace("(",r"\)") + "."
                                                    for tok in ordered_tokens[:-1]
                                                    if tok in (regtokens & tokens))
                                     + ordered_tokens[-1].replace("[",r"\[").replace("(",r"\)")
                                     + "$",  # The rightmost token must always be included 
                                     regkey.lower())]
            if len(rtl_matches) == 0:
                raise KeyError(f"No entry for key '{key}'. Closest matches: "
                               f"{[regkey for regtokens, regkey in matches]}.")
            elif len(rtl_matches) > 1:
                # Problem: a longer path can shadow a shorter one
                # Solution: if there is an exact match, pick that one
                exact_matches = [m for m in rtl_matches if m == key]
                if len(exact_matches) == 1:
                    rtl_matches = exact_matches
                else:
                    # Don’t use KeyError, because then there may be no way do distinguish this
                    # case from a missing case (relied upon by e.g. `get`)
                    raise ValueError("Unable to unambiguously match a registry "
                                     f"entry to '{key}'. Registry contains: {rtl_matches}.")

            return super().__getitem__(rtl_matches[0])

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default




from pathlib import Path
from typing import ClassVar
from pydantic import Field
from valconfig import ValConfig

# These are added to the namespace when deserializing a function
import numpy as np
import math

class Config(ValConfig):
    __default_config_path__: ClassVar=Path(__file__).parent/"defaults.cfg"

    trust_all_inputs: bool=False
    safe_packages: set={"__main__", "scityping"}
    default_namespace: dict=Field(
        default_factory=lambda: {'__builtins__': __builtins__,
                                 'np': np,
                                 'math': math})

config = Config()
